#!/usr/bin/env python
"""
Tests for the ``django-user-tasks`` Celery signal handlers and Django signal.
"""

import logging
from unittest import mock

import pytest
from celery import __version__ as celery_version
from celery import chain, chord, group, shared_task
from packaging import version
from testfixtures import LogCapture

from django.contrib import auth
from django.db import transaction
from django.test import TestCase, TransactionTestCase, override_settings

from user_tasks import user_task_stopped
from user_tasks.models import UserTaskStatus
from user_tasks.signals import start_user_task
from user_tasks.tasks import UserTask

User = auth.get_user_model()


CELERY_VERSION = version.parse(celery_version)
LOGGER = logging.getLogger(__name__)
SIGNAL_LOGGER = 'celery.utils.dispatch.signal'
USER_ID = 1


class SampleTask(UserTask):  # pylint: disable=abstract-method
    """
    Small UserTask subclass for use in test cases.
    """

    @classmethod
    def generate_name(cls, arguments_dict):
        return f"SampleTask: {arguments_dict['argument']}"

    @staticmethod
    def calculate_total_steps(arguments_dict):
        return arguments_dict['kwargs'].get('total_steps', UserTask.calculate_total_steps(arguments_dict))


@shared_task(base=SampleTask, bind=True)
def sample_task(self, user_id, argument, **kwargs):  # pylint: disable=unused-argument
    """
    Example of a specific task inheriting from UserTask.
    """
    print(f'Ran SampleTask for argument "{argument}"')
    return argument


@shared_task(base=SampleTask, bind=True)
def missing_user_id(self, *args, **kwargs):  # pylint: disable=unused-argument
    """
    Example of a UserTask subclass which doesn't specifically require the mandatory user_id argument.
    """


@shared_task(bind=True)
def normal_task(self, *args, **kwargs):  # pylint: disable=unused-argument
    """
    Simple Celery task which doesn't inherit from UserTaskMixin.
    """
    return 'placeholder'


def verify_state(status, eager):
    """
    Assert that completed_steps, state, and attempts are correct for the given value of CELERY_ALWAYS_EAGER.
    """
    assert status.attempts == 1
    if eager:
        assert status.completed_steps == status.total_steps
        assert status.state == UserTaskStatus.SUCCEEDED
    else:
        assert status.completed_steps == 0
        assert status.state == UserTaskStatus.PENDING


SIGNAL_DATA = {}


def receiver(sender, **kwargs):  # pylint: disable=unused-argument
    """
    Signal handler for testing the user_task_stopped signal.
    """
    if 'received_status' in SIGNAL_DATA:
        # Flag duplicate signals
        SIGNAL_DATA['received_status'].state += ' again'
    SIGNAL_DATA['received_status'] = kwargs.get('status')


user_task_stopped.connect(receiver)


class TestCreateUserTask(TestCase):
    """
    Tests of UserTaskStatus creation for new UserTasks.
    """

    def tearDown(self):
        super().tearDown()
        SIGNAL_DATA.clear()

    @classmethod
    def setUpTestData(cls):
        super().setUpTestData()
        cls.user = User.objects.create_user('test_user', 'test@example.com', 'password')

    def test_create_user_task(self):
        """The create_user_task signal handler should create a new UserTaskStatus record"""
        self._create_user_task(eager=False)

    @override_settings(CELERY_ALWAYS_EAGER=True)
    def test_create_user_task_eager(self):
        """Eager tasks should still have UserTaskStatus records created on execution."""
        self._create_user_task(eager=True)
        assert SIGNAL_DATA['received_status'].state == UserTaskStatus.SUCCEEDED

    def test_create_group(self):
        """The create_user_task signal handler should correctly handle celery groups"""
        self._create_group(eager=False)

    def test_create_chain(self):
        """The create_user_task signal handler should correctly handle celery chains."""
        self._create_chain(eager=False)

    def test_create_chord(self):
        """The create_user_task signal handler should correctly handle celery chords"""
        self._create_chord(eager=False)

    def test_create_chord_exclude_body(self):
        """If the body task of a chord is not a UserTask, it should be cleanly omitted from the status."""
        chord([
            sample_task.s(self.user.id, '1', user_task_name='Chord: 1 & 2'),
            sample_task.s(self.user.id, '2', user_task_name='I should be ignored')
        ])(normal_task.s('3'))
        assert UserTaskStatus.objects.count() == 4
        chord_status = UserTaskStatus.objects.get(task_class='celery.chord')
        assert chord_status.task_id
        assert chord_status.parent is None
        assert chord_status.is_container
        assert chord_status.name == 'Chord: 1 & 2'
        assert chord_status.total_steps == 2
        verify_state(chord_status, False)

        group_status = UserTaskStatus.objects.get(task_class='celery.group')
        assert group_status.task_id
        assert group_status.parent_id == chord_status.id
        assert group_status.is_container
        assert group_status.name == 'Chord: 1 & 2'
        assert group_status.total_steps == 2
        verify_state(group_status, False)

        header_tasks = UserTaskStatus.objects.filter(parent=group_status)
        assert len(header_tasks) == 2
        for status in header_tasks:
            assert status.task_id
            assert status.parent_id == group_status.id
            assert not status.is_container
            assert status.name in ['SampleTask: 1', 'SampleTask: 2']
            assert status.total_steps == 1
            verify_state(status, False)

    def test_missing_user_id(self):
        """Queueing of the task should fail if the user ID is not provided."""
        self._verify_type_error(missing_user_id.delay,
                                'Each invocation of a UserTaskMixin subclass must include the user_id')

    def test_invalid_user_id(self):
        """Queueing of the task should fail if an invalid user ID is given."""
        self._verify_type_error(sample_task.delay, 'Invalid user_id: arg1', ('arg1', 'arg2'))

    def test_non_user_task_publish(self):
        """Non-UserTask tasks should still pass through the before_task_publish handler cleanly."""
        normal_task.delay('Argument')
        statuses = UserTaskStatus.objects.all()
        assert not statuses

    @override_settings(CELERY_ALWAYS_EAGER=True, CELERY_IGNORE_RESULT=False)
    def test_non_user_task_success(self):
        """Non-UserTask tasks should still pass through start and success handlers cleanly."""
        result = normal_task.delay('Argument')
        assert result.get() == 'placeholder'
        statuses = UserTaskStatus.objects.all()
        assert not statuses

    def _create_user_task(self, eager):
        """Create a task based on UserTaskMixin and verify some assertions about its corresponding status."""
        result = sample_task.delay(self.user.id, 'Argument')
        statuses = UserTaskStatus.objects.all()
        assert len(statuses) == 1
        status = statuses[0]
        assert status.task_id == result.id
        assert status.task_class == 'test_signals.sample_task'
        assert status.user_id == self.user.id
        assert status.parent is None
        assert not status.is_container
        assert status.name == 'SampleTask: Argument'
        assert status.total_steps == 1
        verify_state(status, eager)

    def _create_chain(self, eager):
        """Create a celery chain and verify some assertions about the corresponding status records"""
        chain(sample_task.si(self.user.id, '1'),
              sample_task.si(self.user.id, '2', user_task_name='Chain: 1, 2, 3'),
              sample_task.si(self.user.id, '3'),
              normal_task.si('Argument')).delay()
        assert UserTaskStatus.objects.count() == 4
        chain_status = UserTaskStatus.objects.get(task_class='celery.chain')
        assert chain_status.task_id
        assert chain_status.parent is None
        assert chain_status.is_container
        assert chain_status.name == 'Chain: 1, 2, 3'
        assert chain_status.total_steps == 3
        verify_state(chain_status, eager)

        children = UserTaskStatus.objects.filter(parent=chain_status)
        assert len(children) == 3
        for status in children:
            assert not status.is_container
            assert status.name in ['SampleTask: 1', 'SampleTask: 2', 'SampleTask: 3']
            assert status.total_steps == 1
            verify_state(status, eager)

    def _create_chord(self, eager):
        """Create a celery chord and verify some assertions about the corresponding status records"""
        chord([
            sample_task.s(self.user.id, '1'),
            sample_task.s(self.user.id, '2', user_task_name='Chord: 1 & 2, then 3')
        ])(sample_task.s(self.user.id, '3'))
        assert UserTaskStatus.objects.count() == 5
        chord_status = UserTaskStatus.objects.get(task_class='celery.chord')
        assert chord_status.task_id
        assert chord_status.parent is None
        assert chord_status.is_container
        assert chord_status.name == 'Chord: 1 & 2, then 3'
        assert chord_status.total_steps == 3
        verify_state(chord_status, eager)

        group_status = UserTaskStatus.objects.get(task_class='celery.group')
        assert group_status.task_id
        assert group_status.parent_id == chord_status.id
        assert group_status.is_container
        assert group_status.name == 'Chord: 1 & 2, then 3'
        assert group_status.total_steps == 2
        verify_state(group_status, eager)

        header_tasks = UserTaskStatus.objects.filter(parent=group_status)
        assert len(header_tasks) == 2
        for status in header_tasks:
            assert status.task_id
            assert status.parent_id == group_status.id
            assert not status.is_container
            assert status.name in ['SampleTask: 1', 'SampleTask: 2']
            assert status.total_steps == 1
            verify_state(status, eager)

        body_status = UserTaskStatus.objects.get(parent=chord_status, is_container=False)
        assert body_status.task_id
        assert body_status.name == 'SampleTask: 3'
        assert body_status.total_steps == 1
        verify_state(body_status, eager)

    def _create_group(self, eager):
        """Create a celery group and verify some assertions about the corresponding status records"""
        result = group(sample_task.s(self.user.id, '1'),
                       sample_task.s(self.user.id, '2', user_task_name='Group: 1, 2')).delay()
        assert UserTaskStatus.objects.count() == 3
        group_status = UserTaskStatus.objects.get(task_class='celery.group')
        assert group_status.task_id == result.id
        assert group_status.parent is None
        assert group_status.is_container
        assert group_status.name == 'Group: 1, 2'
        assert group_status.total_steps == 2
        verify_state(group_status, eager)

        assert len(result.children) == 2
        for result in result.children:
            task_id = result.id
            status = UserTaskStatus.objects.get(task_id=task_id)
            assert status.parent_id == group_status.id
            assert not status.is_container
            assert status.name in ['SampleTask: 1', 'SampleTask: 2']
            assert status.total_steps == 1
            verify_state(status, eager)

    @staticmethod
    def _verify_type_error(func, message, args=()):
        """
        Verify that executing the given function either raises (celery < 4)
        or logs (celery >= 4) a TypeError with the given message.
        """
        if CELERY_VERSION < version.parse('4.0'):
            with pytest.raises(TypeError) as exc_info:
                func(*args)
            assert str(exc_info.value) == message
        else:
            with LogCapture() as capture:
                func(*args)
                assert message in str(capture)


@shared_task(base=SampleTask, bind=True)
def failing_task(self, user_id, argument, **kwargs):
    """
    A task using the UserTask framework which always throws an exception.
    """
    raise Exception('Boom!')


@shared_task(base=SampleTask, bind=True)
def manually_failed_task(self, user_id, argument, **kwargs):  # pylint: disable=unused-argument
    """
    A task using the UserTask framework which fails without throwing an exception.
    """
    self.status.fail("Something went wrong")


@shared_task(bind=True)
def normal_failing_task(self, *args, **kwargs):
    """
    A non-UserTaskMixin Celery task which always throws an exception.
    """
    raise Exception('Boom!')


@shared_task(base=SampleTask, bind=True)
def retried_task(self, user_id, argument, **kwargs):  # pylint: disable=unused-argument
    """
    A task using the UserTask framework which fails once and succeeds on the retry.
    """
    if self.request.retries == 0:
        try:
            raise Exception('Boom!')
        except Exception as exc:
            raise self.retry(exc=exc)
    return argument


@shared_task(bind=True)
def normal_retried_task(self, *args, **kwargs):
    """
    A non-UserTaskMixin Celery task which fails once and succeeds on the retry.
    """
    if self.request.retries == 0:
        try:
            raise Exception('Boom!')
        except Exception as exc:
            raise self.retry(exc=exc)
    return 'placeholder'


@shared_task(base=SampleTask, bind=True)
def self_canceled_task(self, user_id, argument, **kwargs):  # pylint: disable=unused-argument
    """
    This task cancels itself to simulate a task canceled during execution.
    """
    status = self.status
    status.state = UserTaskStatus.CANCELED
    status.save()
    self.status.increment_completed_steps()


class TestStatusChanges(TestCase):
    """
    Tests of signals indicating changes in the task's status.
    """

    def tearDown(self):
        super().tearDown()
        SIGNAL_DATA.clear()

    @classmethod
    def setUpTestData(cls):
        super().setUpTestData()
        cls.user = User.objects.create_user('test_user', 'test@example.com', 'password')

    def test_canceled_before_execution(self):
        """A UserTask which is canceled before execution should have its status updated accordingly."""
        result = sample_task.delay(self.user.id, 'Argument')
        statuses = UserTaskStatus.objects.all()
        assert len(statuses) == 1
        status = statuses[0]
        status.cancel()
        assert status.task_id == result.id
        assert status.task_class == 'test_signals.sample_task'
        assert status.user_id == self.user.id
        assert status.parent is None
        assert not status.is_container
        assert status.name == 'SampleTask: Argument'
        assert status.total_steps == 1
        assert status.completed_steps == 0
        assert status.state == UserTaskStatus.CANCELED
        assert status.attempts == 1
        assert SIGNAL_DATA['received_status'].state == UserTaskStatus.CANCELED

    @override_settings(CELERY_ALWAYS_EAGER=True)
    def test_canceled_during_execution(self):
        """A UserTask which is canceled during execution should have its status updated accordingly."""
        result = self_canceled_task.delay(self.user.id, 'Argument', total_steps=3)
        statuses = UserTaskStatus.objects.all()
        assert len(statuses) == 1
        status = statuses[0]
        assert status.task_id == result.id
        assert status.task_class == 'test_signals.self_canceled_task'
        assert status.user_id == self.user.id
        assert status.parent is None
        assert not status.is_container
        assert status.name == 'SampleTask: Argument'
        assert status.total_steps == 3
        assert status.completed_steps == 1
        assert status.state == UserTaskStatus.CANCELED
        assert status.attempts == 1
        assert SIGNAL_DATA['received_status'].state == UserTaskStatus.CANCELED

    @override_settings(CELERY_ALWAYS_EAGER=True)
    def test_failed(self):
        """A UserTask that failed should have its status updated accordingly."""
        result = failing_task.delay(self.user.id, 'Argument')
        statuses = UserTaskStatus.objects.all()
        assert len(statuses) == 1
        status = statuses[0]
        assert status.task_id == result.id
        assert status.task_class == 'test_signals.failing_task'
        assert status.user_id == self.user.id
        assert status.parent is None
        assert not status.is_container
        assert status.name == 'SampleTask: Argument'
        assert status.total_steps == 1
        assert status.completed_steps == 0
        assert status.state == UserTaskStatus.FAILED
        assert status.attempts == 1
        assert SIGNAL_DATA['received_status'].state == UserTaskStatus.FAILED

    @override_settings(CELERY_ALWAYS_EAGER=True)
    def test_manually_failed(self):
        """A UserTask that failed cleanly (no exception) should have its status updated accordingly."""
        result = manually_failed_task.delay(self.user.id, 'Argument')
        statuses = UserTaskStatus.objects.all()
        assert len(statuses) == 1
        status = statuses[0]
        assert status.task_id == result.id
        assert status.task_class == 'test_signals.manually_failed_task'
        assert status.user_id == self.user.id
        assert status.parent is None
        assert not status.is_container
        assert status.name == 'SampleTask: Argument'
        assert status.total_steps == 1
        assert status.completed_steps == 0
        assert status.state == UserTaskStatus.FAILED
        assert status.attempts == 1
        assert SIGNAL_DATA['received_status'].state == UserTaskStatus.FAILED

    @override_settings(CELERY_ALWAYS_EAGER=True)
    def test_retried(self):
        """A UserTask that is to be retried should have its status updated accordingly."""
        if version.parse('4.2') <= CELERY_VERSION < version.parse('4.4.3'):
            # See https://github.com/celery/celery/issues/4661 for details
            # Should only affect tests, not production (we don't use eager mode in production)
            with mock.patch("celery.app.task.denied_join_result"):
                result = retried_task.delay(self.user.id, 'Argument')
        else:
            result = retried_task.delay(self.user.id, 'Argument')
        statuses = UserTaskStatus.objects.all()
        assert len(statuses) == 1
        status = statuses[0]
        assert status.task_id == result.id
        assert status.task_class == 'test_signals.retried_task'
        assert status.user_id == self.user.id
        assert status.parent is None
        assert not status.is_container
        assert status.name == 'SampleTask: Argument'
        if CELERY_VERSION < version.parse('4.4.3'):
            # Until the fix in 4.4.3, eager task execution sent the retry signal last
            assert status.state == UserTaskStatus.RETRYING
        else:
            assert status.state == UserTaskStatus.SUCCEEDED
        assert status.total_steps == 1
        assert status.completed_steps == 1
        assert status.attempts == 2

    @override_settings(CELERY_ALWAYS_EAGER=True)
    def test_non_user_task_success(self):
        """Non-UserTask tasks should still pass through the failure handler cleanly."""
        normal_failing_task.delay('Argument')
        statuses = UserTaskStatus.objects.all()
        assert not statuses
        assert 'received_status' not in SIGNAL_DATA

    @override_settings(CELERY_ALWAYS_EAGER=True)
    def test_non_user_task_retry(self):
        """Non-UserTask tasks should still pass through the retry handler cleanly."""
        normal_retried_task.delay('Argument')
        statuses = UserTaskStatus.objects.all()
        assert not statuses
        assert 'received_status' not in SIGNAL_DATA

    def test_duplicate_stopped_signal(self):
        """The test signal receiver should be able to detect and flag duplicate signals"""
        sample_task.delay(self.user.id, 'Argument')
        status = UserTaskStatus.objects.all()[0]
        user_task_stopped.send_robust(sender=UserTaskStatus, status=status)
        status.cancel()
        user_task_stopped.send_robust(sender=UserTaskStatus, status=status)
        assert SIGNAL_DATA['received_status'].state == UserTaskStatus.CANCELED + ' again'


class TestConnectionClosing(TransactionTestCase):
    """
    Tests the logic around closing obsolete connections during task start.
    """
    @mock.patch('user_tasks.signals.close_old_connections', autospec=True)
    def test_connections_closed_outside_atomic_block(self, mock_close_old_connections):
        start_user_task(sender=SampleTask)
        mock_close_old_connections.assert_called_once_with()

    @mock.patch('user_tasks.signals.close_old_connections', autospec=True)
    def test_connections_not_closed_inside_atomic_block(self, mock_close_old_connections):
        with transaction.atomic():
            start_user_task(sender=SampleTask)
            assert mock_close_old_connections.called is False

    @mock.patch('user_tasks.signals.close_old_connections', autospec=True)
    def test_connections_not_closed_when_we_cant_get_a_connection(self, mock_close_old_connections):
        with mock.patch('user_tasks.signals.transaction.get_connection', side_effect=Exception):
            start_user_task(sender=SampleTask)
            assert mock_close_old_connections.called is False



#!/usr/bin/env python
"""
Tests for the user_tasks subclasses of celery.Task.
"""

import logging
from datetime import timedelta
from uuid import uuid4

from celery import Task, shared_task

from django.contrib import auth
from django.test import TestCase, override_settings
from django.utils.timezone import now

from user_tasks.models import UserTaskArtifact, UserTaskStatus
from user_tasks.tasks import UserTask, UserTaskMixin, purge_old_user_tasks

User = auth.get_user_model()

LOGGER = logging.getLogger(__name__)


class MinimalTask(Task, UserTaskMixin):  # pylint: disable=abstract-method
    """
    A minimal UserTaskMixin subclass which uses the default name and total_steps generation.
    """


@shared_task(base=MinimalTask, bind=True)
def minimal_task(self, user_id, argument, **kwargs):  # pylint: disable=unused-argument
    """
    Example usage of the MinimalTask class.
    """
    return argument


class SampleTask(UserTask):  # pylint: disable=abstract-method
    """
    Small UserTask subclass for use in test cases.
    """

    @classmethod
    def generate_name(cls, arguments_dict):
        arg1 = arguments_dict['arg1']
        arg2 = arguments_dict['arg2']
        return f'SampleTask: {arg1}, {arg2}'

    @staticmethod
    def calculate_total_steps(arguments_dict):
        arg1 = arguments_dict['arg1']
        arg2 = arguments_dict['arg2']
        return arg1 * arg2


@shared_task(base=SampleTask, bind=True)
def sample_task(self, user_id, arg1, arg2, **kwargs):  # pylint: disable=unused-argument
    """
    Example of a specific task inheriting from UserTask.
    """
    return arg1, arg2


class TestUserTasks(TestCase):
    """
    Tests of UserTaskMixin and UserTask.
    """

    @classmethod
    def setUpTestData(cls):
        super().setUpTestData()
        cls.user = User.objects.create_user('test_user', 'test@example.com', 'password')

    def test_generate_name_omitted(self):
        """Using UserTaskMixin without generate_name() implemented should default to the task function name."""
        minimal_task.delay(self.user.id, 'Argument')
        statuses = UserTaskStatus.objects.all()
        assert len(statuses) == 1
        assert statuses[0].name == 'minimal_task'

    def test_generate_name(self):
        """generate_name() should be able to generate a task name based on the task's arguments."""
        result = sample_task.delay(self.user.id, 1, 2)
        status = UserTaskStatus.objects.get(task_id=result.id)
        assert status.name == 'SampleTask: 1, 2'

        result = sample_task.delay(self.user.id, 3, arg2=4)
        status = UserTaskStatus.objects.get(task_id=result.id)
        assert status.name == 'SampleTask: 3, 4'

        result = sample_task.delay(arg2=6, arg1=5, user_id=self.user.id)
        status = UserTaskStatus.objects.get(task_id=result.id)
        assert status.name == 'SampleTask: 5, 6'

    def test_calculate_total_steps(self):
        """calculate_total_steps() should generate the total_steps value based on the task's arguments."""
        result = sample_task.delay(self.user.id, 1, 2)
        status = UserTaskStatus.objects.get(task_id=result.id)
        assert status.total_steps == 2

        result = sample_task.delay(self.user.id, 3, arg2=4)
        status = UserTaskStatus.objects.get(task_id=result.id)
        assert status.total_steps == 12

        result = sample_task.delay(arg2=6, arg1=5, user_id=self.user.id)
        status = UserTaskStatus.objects.get(task_id=result.id)
        assert status.total_steps == 30


@override_settings(CELERY_ALWAYS_EAGER=True)
class TestPurgeOldUserTasks(TestCase):
    """
    Tests of the Celery cleanup task for old user task database records.
    """

    @classmethod
    def setUpTestData(cls):
        super().setUpTestData()
        cls.user = User.objects.create_user('test_user', 'test@example.com', 'password')

    def test_old_data(self):
        """The cleanup task should purge old database records."""
        self._create_records(now() - timedelta(days=31))
        purge_old_user_tasks.delay()
        assert UserTaskStatus.objects.count() == 0
        assert UserTaskArtifact.objects.count() == 0

    def test_recent_data(self):
        """The cleanup task should leave recent database records intact."""
        self._create_records(now() - timedelta(days=29))
        purge_old_user_tasks.delay()
        assert UserTaskStatus.objects.count() == 1
        assert UserTaskArtifact.objects.count() == 1

    def test_mixed_data(self):
        """The cleanup task should purge only old database records."""
        self._create_records(now() - timedelta(days=29))
        status = UserTaskStatus.objects.all()[0]
        self._create_records(now() - timedelta(days=31))
        purge_old_user_tasks.delay()
        assert UserTaskStatus.objects.count() == 1
        assert UserTaskArtifact.objects.count() == 1
        assert UserTaskStatus.objects.filter(pk=status.id).exists()

    def _create_records(self, created):
        """
        Create a UserTaskStatus and UserTaskArtifact with the specified creation date.
        """
        data = {
            'name': 'SampleTask', 'state': UserTaskStatus.SUCCEEDED, 'task_class': 'test_tasks.sample_task',
            'task_id': str(uuid4()), 'total_steps': 5, 'user': self.user
        }
        status = UserTaskStatus.objects.create(**data)
        completed = created + timedelta(hours=1)
        UserTaskStatus.objects.filter(pk=status.id).update(created=created, modified=completed)
        artifact = UserTaskArtifact.objects.create(status=status, text='Lorem ipsum')
        UserTaskArtifact.objects.filter(pk=artifact.id).update(created=completed, modified=completed)



#!/usr/bin/env python
"""
Tests for the ``django-user-tasks`` REST API.
"""

import logging
from datetime import timedelta
from unittest import mock
from uuid import uuid4

import rules

from django.contrib import auth
from django.urls import reverse
from django.utils.timezone import now

from rest_framework.test import APITestCase

from user_tasks.models import UserTaskArtifact, UserTaskStatus
from user_tasks.rules import add_rules
from user_tasks.serializers import ArtifactSerializer, StatusSerializer

User = auth.get_user_model()

LOGGER = logging.getLogger(__name__)


# Helper functions for stuff that pylint complains about without disable comments

def _context(response):
    """
    Get a context dictionary for a serializer appropriate for the given response.
    """
    return {'request': response.wsgi_request}


def _data(response):
    """
    Get the serialized data dictionary from the given REST API test response.
    """
    return response.data


class TestRestApi(APITestCase):
    """
    Tests of the REST API calls.
    """

    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        add_rules()

    @classmethod
    def tearDownClass(cls):
        super().tearDownClass()
        rules.rulesets.default_rules.clear()

    @classmethod
    def setUpTestData(cls):
        super().setUpTestData()
        cls.user = User.objects.create_user('test_user', 'test@example.com', 'password')
        cls.superuser = User.objects.create_superuser('admin', 'admin@example.com', 'password')
        cls.other_user = User.objects.create_user('other_user', 'other@example.com', 'password')
        cls.status = UserTaskStatus.objects.create(
            user=cls.user, task_id=str(uuid4()), task_class='test_rest_api.sample_task', name='SampleTask 2',
            total_steps=5)
        cls.artifact = UserTaskArtifact.objects.create(status=cls.status, text='Lorem ipsum')
        cls.older_status = UserTaskStatus.objects.create(
            user=cls.user, task_id=str(uuid4()), task_class='test_rest_api.sample_task', name='SampleTask 1',
            total_steps=5)
        yesterday = now() - timedelta(days=1)
        UserTaskStatus.objects.filter(pk=cls.older_status.id).update(created=yesterday, modified=yesterday)

    def setUp(self):
        super().setUp()
        self.status.refresh_from_db()
        self.older_status.refresh_from_db()

    def test_artifact_detail(self):
        """Users should be able to access artifacts for tasks they triggered."""
        self._login(self.user)
        response = self.client.get(reverse('usertaskartifact-detail', args=[self.artifact.uuid]))
        assert response.status_code == 200
        serializer = ArtifactSerializer(self.artifact, context=_context(response))
        assert _data(response) == serializer.data

    def test_artifact_detail_anonymous(self):
        """Anonymous users should be unable to access task artifacts via the API."""
        response = self.client.get(reverse('usertaskartifact-detail', args=[self.artifact.uuid]))
        assert response.status_code == 403

    def test_artifact_detail_superuser(self):
        """Superusers should be able to access artifacts for all tasks."""
        self._login(self.superuser)
        response = self.client.get(reverse('usertaskartifact-detail', args=[self.artifact.uuid]))
        assert response.status_code == 200
        serializer = ArtifactSerializer(self.artifact, context=_context(response))
        assert _data(response) == serializer.data

    def test_artifact_detail_other_user(self):
        """Users should be unable to access artifacts for another user's tasks via the API."""
        self._login(self.other_user)
        response = self.client.get(reverse('usertaskartifact-detail', args=[self.artifact.uuid]))
        assert response.status_code == 404

    def test_artifact_list(self):
        """Users should be able to access a list of their tasks' artifacts."""
        self._login(self.user)
        response = self.client.get(reverse('usertaskartifact-list'))
        assert response.status_code == 200
        serializer = ArtifactSerializer(self.artifact, context=_context(response))
        assert _data(response) == [serializer.data]

    def test_artifact_list_anonymous(self):
        """Anonymous should be unable to access lists of task artifacts."""
        response = self.client.get(reverse('usertaskartifact-list'))
        assert response.status_code == 403

    def test_artifact_list_superuser(self):
        """Superusers should be able to access a list of all artifacts."""
        self._login(self.superuser)
        response = self.client.get(reverse('usertaskartifact-list'))
        assert response.status_code == 200
        serializer = ArtifactSerializer(self.artifact, context=_context(response))
        assert _data(response) == [serializer.data]

    def test_artifact_list_other_user(self):
        """Users should be unable to see artifacts for other users in their task artifact list."""
        self._login(self.other_user)
        response = self.client.get(reverse('usertaskartifact-list'))
        assert response.status_code == 200
        assert _data(response) == []

    def test_status_cancel(self):
        """Users should be able to cancel tasks they no longer wish to complete."""
        self._login(self.user)
        response = self.client.post(reverse('usertaskstatus-cancel', args=[self.status.uuid]))
        assert response.status_code == 200
        self.status.refresh_from_db()
        assert self.status.state == UserTaskStatus.CANCELED

    def test_status_cancel_anonymous(self):
        """Anonymous users should be unable to cancel tasks."""
        response = self.client.post(reverse('usertaskstatus-cancel', args=[self.status.uuid]))
        assert response.status_code == 403
        self.status.refresh_from_db()
        assert self.status.state == UserTaskStatus.PENDING

    def test_status_cancel_superuser(self):
        """Superusers should be able to cancel any task."""
        self._login(self.superuser)
        response = self.client.post(reverse('usertaskstatus-cancel', args=[self.status.uuid]))
        assert response.status_code == 200
        self.status.refresh_from_db()
        assert self.status.state == UserTaskStatus.CANCELED

    def test_status_cancel_other_user(self):
        """Users should be unable to cancel the tasks of other users."""
        self._login(self.other_user)
        response = self.client.post(reverse('usertaskstatus-cancel', args=[self.status.uuid]))
        assert response.status_code == 404
        self.status.refresh_from_db()
        assert self.status.state == UserTaskStatus.PENDING

    @mock.patch('django.contrib.auth.models.User.has_perm')
    def test_status_cancel_view_only(self, mock_has_perm):
        """A user with view but not cancel permission on a status record should be unable to cancel it."""
        def no_cancel(permission, status):
            """Deny user_tasks.cancel_usertaskstatus permission for any particular status record."""
            if permission == 'user_tasks.cancel_usertaskstatus':
                return status is None
            return True
        mock_has_perm.side_effect = no_cancel
        self._login(self.user)
        response = self.client.post(reverse('usertaskstatus-cancel', args=[self.status.uuid]))
        assert response.status_code == 403
        self.status.refresh_from_db()
        assert self.status.state == UserTaskStatus.PENDING

    def test_status_delete(self):
        """Users should be able to delete their own task status records when they're done with them."""
        self._login(self.user)
        response = self.client.delete(reverse('usertaskstatus-detail', args=[self.status.uuid]))
        assert response.status_code == 204
        assert not UserTaskStatus.objects.filter(pk=self.status.id).exists()

    def test_status_delete_anonymous(self):
        """Anonymous users should not be able to delete task status records."""
        response = self.client.delete(reverse('usertaskstatus-detail', args=[self.status.uuid]))
        assert response.status_code == 403
        assert UserTaskStatus.objects.filter(pk=self.status.id).exists()

    def test_status_delete_superuser(self):
        """Users should be able to delete their own task status records when they're done with them."""
        self._login(self.superuser)
        response = self.client.delete(reverse('usertaskstatus-detail', args=[self.status.uuid]))
        assert response.status_code == 204
        assert not UserTaskStatus.objects.filter(pk=self.status.id).exists()

    def test_status_delete_other_user(self):
        """Users should not be able to delete the task status records of other users."""
        self._login(self.other_user)
        response = self.client.delete(reverse('usertaskstatus-detail', args=[self.status.uuid]))
        assert response.status_code == 404
        assert UserTaskStatus.objects.filter(pk=self.status.id).exists()

    def test_status_detail(self):
        """Users should be able to access status records for tasks they triggered."""
        self._login(self.user)
        response = self.client.get(reverse('usertaskstatus-detail', args=[self.status.uuid]))
        assert response.status_code == 200
        serializer = StatusSerializer(self.status, context=_context(response))
        assert _data(response) == serializer.data

    def test_status_detail_anonymous(self):
        """Anonymous users should be unable to access task status records via the API."""
        response = self.client.get(reverse('usertaskstatus-detail', args=[self.status.uuid]))
        assert response.status_code == 403

    def test_status_detail_superuser(self):
        """Superusers should be able to access all status records."""
        self._login(self.superuser)
        response = self.client.get(reverse('usertaskstatus-detail', args=[self.status.uuid]))
        assert response.status_code == 200
        serializer = StatusSerializer(self.status, context=_context(response))
        assert _data(response) == serializer.data

    def test_status_detail_other_user(self):
        """Users should be unable to access status records for another user's tasks via the API."""
        self._login(self.other_user)
        response = self.client.get(reverse('usertaskstatus-detail', args=[self.status.uuid]))
        assert response.status_code == 404

    def test_status_list(self):
        """Users should be able to access a list of their tasks' status records."""
        self._login(self.user)
        response = self.client.get(reverse('usertaskstatus-list'))
        assert response.status_code == 200
        serializer = StatusSerializer(
            [self.status, self.older_status], context=_context(response), many=True)
        assert _data(response) == serializer.data

    def test_status_list_anonymous(self):
        """Anonymous should be unable to access lists of task status records."""
        response = self.client.get(reverse('usertaskstatus-list'))
        assert response.status_code == 403

    def test_status_list_superuser(self):
        """Superusers should be able to access a list of all task status records."""
        self._login(self.superuser)
        response = self.client.get(reverse('usertaskstatus-list'))
        assert response.status_code == 200
        serializer = StatusSerializer(
            [self.status, self.older_status], context=_context(response), many=True)
        assert _data(response) == serializer.data

    def test_status_list_other_user(self):
        """Users should be unable to see tasks for other users in their task status list."""
        self._login(self.other_user)
        response = self.client.get(reverse('usertaskstatus-list'))
        assert response.status_code == 200
        assert _data(response) == []

    def _login(self, user):
        """
        Log the test client in as the specified user.
        """
        self.client.force_authenticate(user)



#!/usr/bin/env python
"""
Tests for the REST API model serializers.
"""

import shutil
from uuid import uuid4

from django.conf import settings
from django.contrib import auth
from django.core.files.uploadedfile import SimpleUploadedFile
from django.test import TestCase
from django.urls import reverse
from django.utils.timezone import localtime

from rest_framework.test import APIRequestFactory

from user_tasks.models import UserTaskArtifact, UserTaskStatus
from user_tasks.serializers import ArtifactSerializer, StatusSerializer

User = auth.get_user_model()


def _format(datetime):
    """
    Generate the same text representation of the given datetime that DRF would.
    """
    return localtime(datetime).isoformat()


class TestStatusSerializer(TestCase):
    """
    Tests of the serializer for UserTaskStatus model instances.
    """

    @classmethod
    def setUpTestData(cls):
        super().setUpTestData()
        cls.user = User.objects.create_user('test_user', 'test@example.com', 'password')

    def test_output(self):
        """The serializer should generate the expected data for a status record."""
        status = UserTaskStatus.objects.create(user=self.user, task_id=str(uuid4()), name='SampleTask', total_steps=4)
        expected = {
            'name': 'SampleTask',
            'state': 'Pending',
            'state_text': 'Pending',
            'completed_steps': 0,
            'total_steps': 4,
            'attempts': 1,
            'created': _format(status.created),
            'modified': _format(status.modified),
            'artifacts': []
        }
        serializer = StatusSerializer(status)
        assert serializer.data == expected

    def test_output_with_artifact(self):
        """The serializer should include URLs for any generated artifacts."""
        status = UserTaskStatus.objects.create(user=self.user, task_id=str(uuid4()), name='SampleTask', total_steps=4)
        artifact = UserTaskArtifact.objects.create(status=status, text='Lorem Ipsum')
        expected = {
            'name': 'SampleTask',
            'state': 'Pending',
            'state_text': 'Pending',
            'completed_steps': 0,
            'total_steps': 4,
            'attempts': 1,
            'created': _format(status.created),
            'modified': _format(status.modified),
            'artifacts': [f'http://testserver/artifacts/{artifact.uuid}/']
        }
        request = APIRequestFactory().get(reverse('usertaskstatus-detail', args=[status.uuid]))
        serializer = StatusSerializer(status, context={'request': request})
        assert serializer.data == expected


class TestArtifactSerializer(TestCase):
    """
    Tests of the serializer for UserTaskArtifact model instances
    """

    @classmethod
    def setUpTestData(cls):
        super().setUpTestData()
        user = User.objects.create_user('test_user', 'test@example.com', 'password')
        cls.status = UserTaskStatus.objects.create(user=user, task_id=str(uuid4()), name='SampleTask', total_steps=4)

    @classmethod
    def tearDownClass(cls):
        super().tearDownClass()
        # Clean up temp files
        shutil.rmtree(settings.MEDIA_ROOT)

    def test_file(self):
        """The serializer should handle file artifacts correctly."""
        uploaded_file = SimpleUploadedFile('filename.txt', b'Content of the file')
        artifact = UserTaskArtifact.objects.create(status=self.status, file=uploaded_file)
        expected = {
            'status': f'http://testserver/tasks/{self.status.uuid}/',
            'name': 'Output',
            'created': _format(artifact.created),
            'modified': _format(artifact.modified),
            'file': artifact.file.url,
            'text': '',
            'url': '',
        }
        request = APIRequestFactory().get(reverse('usertaskartifact-detail', args=[artifact.uuid]))
        serializer = ArtifactSerializer(artifact, context={'request': request})
        assert serializer.data == expected

    def test_text(self):
        """The serializer should handle text block artifacts correctly."""
        artifact = UserTaskArtifact.objects.create(status=self.status, text='Got your output right here.')
        expected = {
            'status': f'http://testserver/tasks/{self.status.uuid}/',
            'name': 'Output',
            'created': _format(artifact.created),
            'modified': _format(artifact.modified),
            'file': '',
            'text': artifact.text,
            'url': '',
        }
        request = APIRequestFactory().get(reverse('usertaskartifact-detail', args=[artifact.uuid]))
        serializer = ArtifactSerializer(artifact, context={'request': request})
        assert serializer.data == expected

    def test_url(self):
        """The serializer should handle URL artifacts correctly."""
        artifact = UserTaskArtifact.objects.create(status=self.status, url='http://www.example.com/output/3/')
        expected = {
            'status': f'http://testserver/tasks/{self.status.uuid}/',
            'name': 'Output',
            'created': _format(artifact.created),
            'modified': _format(artifact.modified),
            'file': '',
            'text': '',
            'url': artifact.url,
        }
        request = APIRequestFactory().get(reverse('usertaskartifact-detail', args=[artifact.uuid]))
        serializer = ArtifactSerializer(artifact, context={'request': request})
        assert serializer.data == expected



"""
Django admin configuration for the ``django-user-tasks`` models.
"""

from django.contrib import admin

from .models import UserTaskArtifact, UserTaskStatus


@admin.register(UserTaskArtifact)
class UserTaskArtifactAdmin(admin.ModelAdmin):
    """
    Configuration for UserTaskArtifact admin panel.
    """

    list_display = ('created', 'uuid', 'status', 'name', 'text')
    list_filter = ('name',)
    ordering = ('-created',)
    search_fields = ('uuid', 'name', 'text')
    raw_id_fields = ('status',)


@admin.register(UserTaskStatus)
class UserTaskStatusAdmin(admin.ModelAdmin):
    """
    Configuration for UserTaskStatus admin panel.
    """

    list_display = ('created', 'uuid', 'state', 'user', 'name')
    list_filter = ('state',)
    ordering = ('-created',)
    search_fields = (
        'uuid', 'task_id', 'task_class', 'name', 'user__username', 'user__email'
    )
    readonly_fields = ('parent', )



"""
URLs for user_tasks.
"""

from rest_framework.routers import SimpleRouter

from user_tasks.views import ArtifactViewSet, StatusViewSet

ROUTER = SimpleRouter()
ROUTER.register(r'artifacts', ArtifactViewSet, basename='usertaskartifact')
ROUTER.register(r'tasks', StatusViewSet, basename='usertaskstatus')

urlpatterns = ROUTER.urls



"""
Management of user-triggered asynchronous tasks in Django projects.
"""

from django.dispatch import Signal

__version__ = '3.2.0'


# This signal is emitted when a user task reaches any final state:
# SUCCEEDED, FAILED, or CANCELED
# providing_args = ['status']
user_task_stopped = Signal()



#!/usr/bin/env python
"""
Tests for the ``django-user-tasks`` admin module.
"""

from django.contrib import auth
from django.test import TestCase
from django.urls import reverse

User = auth.get_user_model()


class AdminTestCase(TestCase):
    """
    Tests for the ``user_tasks`` application's Django admin pages.
    """
    def setUp(self):
        super().setUp()
        self.user = User.objects.create_user(
            username='tester',
            email='tester@example.com',
            password='password',
            is_staff=True,
            is_superuser=True,
        )
        self.client.login(username=self.user.username, password='password')

    def test_artifact_list(self):
        """
        Make sure the main UserTaskArtifact admin page loads.
        """
        response = self.client.get(reverse('admin:user_tasks_usertaskartifact_changelist'))
        assert response.status_code == 200

    def test_status_list(self):
        """
        Make sure the main UserTaskStatus admin page loads.
        """
        response = self.client.get(reverse('admin:user_tasks_usertaskstatus_changelist'))
        assert response.status_code == 200



#!/usr/bin/env python
"""
Tests for the `django-user-tasks` models module.
"""

import logging
from uuid import uuid4

import pytest

from django.contrib import auth
from django.core.files.uploadedfile import SimpleUploadedFile
from django.test import TestCase

from user_tasks.exceptions import TaskCanceledException
from user_tasks.models import UserTaskArtifact, UserTaskStatus

User = auth.get_user_model()

LOGGER = logging.getLogger(__name__)


class TestUserTaskStatus(TestCase):
    """
    Tests of the UserTask model.
    """

    @classmethod
    def setUpTestData(cls):
        super().setUpTestData()
        cls.user = User.objects.create_user('test_user', 'test@example.com', 'password')

    def test_cancel_parent(self):
        """Canceling a container status should also cancel its children."""
        parent = self._status(is_container=True, state=UserTaskStatus.PENDING)
        status = self._status(parent=parent, state=UserTaskStatus.PENDING)
        parent.cancel()
        status.refresh_from_db()
        assert parent.state == UserTaskStatus.CANCELED
        assert status.state == UserTaskStatus.CANCELED

    def test_cancel_finished_task(self):
        """Attempting to cancel an already-finished task should have no effect."""
        status = self._status(state=UserTaskStatus.SUCCEEDED)
        status.cancel()
        assert status.state == UserTaskStatus.SUCCEEDED

    def test_fail_with_parent(self):
        """A task which fails should also mark any parent as having failed."""
        parent = self._status(is_container=True)
        status = self._status(parent=parent)
        status.fail('Oops!')
        parent.refresh_from_db()
        assert status.state == UserTaskStatus.FAILED
        assert parent.state == UserTaskStatus.FAILED
        artifact = UserTaskArtifact.objects.get(status=status)
        assert artifact.text == 'Oops!'
        parent_artifact = UserTaskArtifact.objects.get(status=parent)
        assert parent_artifact.text == 'Oops!'

    def test_increment_completed_steps(self):
        """increment_completed_steps() should update both the current status and any parent status."""
        parent = self._status(is_container=True, total_steps=0)
        status = self._status(parent=parent, total_steps=2)
        status.increment_completed_steps()
        parent.refresh_from_db()
        assert status.completed_steps == 1
        assert parent.completed_steps == 1

    def test_increment_completed_steps_explicit_steps(self):
        """increment_completed_steps() should support explicit step counts greater than 1."""
        status = self._status(total_steps=5)
        status.increment_completed_steps(3)
        assert status.completed_steps == 3
        status.refresh_from_db()
        assert status.completed_steps == 3

    def test_increment_total_steps(self):
        """increment_total_steps() should correctly update the status and any parent."""
        parent = self._status(is_container=True, total_steps=3)
        self._status(parent=parent, total_steps=2)
        status = self._status(parent=parent, total_steps=1)
        status.increment_total_steps(2)
        parent.refresh_from_db()
        assert status.total_steps == 3
        assert parent.total_steps == 5

    def test_retry(self):
        """retry() should update the status fields accordingly and not impact any parent."""
        parent = self._status(is_container=True)
        status = self._status(parent=parent)
        status.retry()
        parent.refresh_from_db()
        assert status.state == UserTaskStatus.RETRYING
        assert status.attempts == 2
        assert parent.state == UserTaskStatus.IN_PROGRESS

    def test_set_name(self):
        """set_name() should update both the subject and any parent."""
        parent = self._status(is_container=True)
        status = self._status(parent=parent)
        status.set_name('New Name')
        parent.refresh_from_db()
        assert status.name == 'New Name'
        assert parent.name == 'New Name'

    def test_set_state(self):
        """It should be possible to set a custom state for a UserTaskStatus."""
        status = self._status()
        status.set_state('Custom State')
        assert status.state == 'Custom State'
        status.refresh_from_db()
        assert status.state == 'Custom State'

    def test_set_state_canceled(self):
        """Setting a custom state for a canceled task should fail."""
        status = self._status(state=UserTaskStatus.CANCELED)
        with pytest.raises(TaskCanceledException):
            status.set_state('Custom State')
        assert status.state == UserTaskStatus.CANCELED
        status.refresh_from_db()
        assert status.state == UserTaskStatus.CANCELED

    def test_set_state_parent(self):
        """Setting a custom state should propagate to the state's parent."""
        parent = self._status(is_container=True, task_class='celery.chain')
        status = self._status(parent=parent)
        status.set_state('Custom State')
        assert status.state == 'Custom State'
        status.refresh_from_db()
        assert status.state == 'Custom State'
        assert parent.state == 'Custom State'
        parent.refresh_from_db()
        assert parent.state == 'Custom State'

    def test_start(self):
        """The start() method should update the status record appropriately."""
        status = self._status(state=UserTaskStatus.PENDING)
        assert status.state == UserTaskStatus.PENDING
        status.start()
        assert status.state == UserTaskStatus.IN_PROGRESS
        status.refresh_from_db()
        assert status.state == UserTaskStatus.IN_PROGRESS

    def test_start_canceled_task(self):
        """Attempting to start a canceled task should raise a TaskCanceledException."""
        status = self._status(state=UserTaskStatus.CANCELED)
        with pytest.raises(TaskCanceledException):
            status.start()
        assert status.state == UserTaskStatus.CANCELED

    def test_start_with_failed_sibling(self):
        """If a sibling task has already failed, the parent status should not change when the task starts."""
        parent = self._status(is_container=True, state=UserTaskStatus.FAILED)
        self._status(parent=parent, state=UserTaskStatus.FAILED)
        child2 = self._status(parent=parent, state=UserTaskStatus.PENDING)
        child2.start()
        parent.refresh_from_db()
        assert child2.state == UserTaskStatus.IN_PROGRESS
        assert parent.state == UserTaskStatus.FAILED

    def test_start_with_parent(self):
        """Starting a pending task should also update any parent task status."""
        parent = self._status(is_container=True, state=UserTaskStatus.PENDING)
        status = self._status(parent=parent, state=UserTaskStatus.PENDING)
        status.start()
        assert status.state == UserTaskStatus.IN_PROGRESS
        parent.refresh_from_db()
        assert parent.state == UserTaskStatus.IN_PROGRESS

    def test_string_representation(self):
        """UserTaskStatus instances should have a reasonable string representation."""
        status = self._status()
        assert str(status) == '<UserTaskStatus: SampleTask>'

    def test_succeed(self):
        """succeed() should update its parent only if all its other children have already succeeded."""
        parent = self._status(is_container=True, total_steps=10)
        child1 = self._status(parent=parent)
        child2 = self._status(parent=parent)
        child1.succeed()
        parent.refresh_from_db()
        assert child1.state == UserTaskStatus.SUCCEEDED
        assert child1.completed_steps == child1.total_steps
        assert parent.state == UserTaskStatus.IN_PROGRESS
        assert parent.completed_steps == child1.total_steps
        child2.succeed()
        parent.refresh_from_db()
        assert child2.state == UserTaskStatus.SUCCEEDED
        assert child2.completed_steps == child2.total_steps
        assert parent.state == UserTaskStatus.SUCCEEDED
        assert parent.completed_steps == child1.total_steps + child2.total_steps

    def _status(self, **kwargs):
        """Generate a sample UserTaskStatus instance, optionally overriding fields with keyword arguments."""
        data = {
            'name': 'SampleTask', 'state': UserTaskStatus.IN_PROGRESS, 'task_class': 'test_models.sample_task',
            'task_id': str(uuid4()), 'total_steps': 5, 'user': self.user}
        data.update(kwargs)
        return UserTaskStatus.objects.create(**data)


class TestUserTaskArtifact(TestCase):
    """
    Tests of the UserTaskArtifact model.
    """

    @classmethod
    def setUpTestData(cls):
        super().setUpTestData()
        cls.user = User.objects.create_user('test_user', 'test@example.com', 'password')

    def test_file_string(self):
        """File-based UserTaskArtifacts should have a reasonable string representation"""
        status = self._status()
        uploaded_file = SimpleUploadedFile('file.txt', b'File content')
        artifact = UserTaskArtifact(status=status, file=uploaded_file)
        assert str(artifact) == '<UserTaskArtifact: (Output) file.txt>'

    def test_text_string(self):
        """Text-based UserTaskArtifacts should have a reasonable string representation."""
        status = self._status()
        artifact = UserTaskArtifact(status=status, text='This was the output text.')
        assert str(artifact) == '<UserTaskArtifact: (Output) This was the output text.>'

    def test_long_text_string(self):
        """Long text artifacts should be cropped for the string representation."""
        status = self._status()
        text = 'blah ' * 50
        artifact = UserTaskArtifact(status=status, name='Cropped', text=text)
        assert str(artifact) == f"<UserTaskArtifact: (Cropped) {text[:47]}...>"

    def test_url_string(self):
        """URL-based UserTaskArtifacts should have a reasonable string representation."""
        status = self._status()
        artifact = UserTaskArtifact(status=status, url='http://www.edx.org/')
        assert str(artifact) == '<UserTaskArtifact: (Output) http://www.edx.org/>'

    def _status(self, **kwargs):
        """Generate a sample UserTaskStatus instance, optionally overriding fields with keyword arguments."""
        data = {
            'name': 'SampleTask', 'state': UserTaskStatus.IN_PROGRESS, 'task_class': 'test_models.sample_task',
            'task_id': str(uuid4()), 'total_steps': 5, 'user': self.user}
        data.update(kwargs)
        return UserTaskStatus.objects.create(**data)
